#ident "@(#) $Id: lpfwrite.c,v 1.12 2002/10/13 16:05:21 bzfkocht Exp $"
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*   File....: lpfwrite.c                                                    */
/*   Name....: LP Format File Writer                                         */
/*   Author..: Thorsten Koch                                                 */
/*   Copyright by Author, All rights reserved                                */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*
 * Copyright (C) 2001 by Thorsten Koch <koch@zib.de>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "mshell.h"
#include "portab.h"
#include "mme.h"
#include "rbt.h"
#include "lpstore.h"

/* A specification for the LP file format can be found in the
 * ILOG CPLEX 7.0 Reference Manual page 527.
 */
void lpf_write(
   const Lps* lp,
   FILE*      fp)
{
   const Var* var;
   const Con* con;
   const Nzo* nzo;
   Bool  first;
   Bool  separate = FALSE;
   int   cnt;
   char  tmp[MPS_NAME_LEN + 1];
   int   i;
   
   assert(lp     != NULL);
   assert(fp     != NULL);
   
   fprintf(fp, "\\This file was automatically generated by zimpl\n");
   fprintf(fp, "\\Problem name: %s\n", lp->name);   
   fprintf(fp, "%s\n", (lp->direct == LP_MIN) ? "Minimize" : "Maximize");
   fprintf(fp, " %s: ", lp->objname == NULL ? "Objective" : lp->objname);
   
   for(var = lp->var_root, cnt = 0; var != NULL; var = var->next)
   {
      /* If cost is zero, do not include in objective function
       */
      if (fabs(var->cost) < EPS_ZERO)
         continue;

      /* If the variable is free and when minimizing the lower or
       * when maximizing the upper bound is zero,
       * it can be discarded.
       */
      if ((var->size == 0) && (fabs((lp->direct == LP_MIN) ?
         var->lower : var->upper) < EPS_ZERO))
         continue;
      
      lps_makename(tmp, MPS_NAME_LEN + 1, var->name, var->number);

      if (EQ(var->cost, 1.0))
         fprintf(fp, " + %s", tmp);
      else if (EQ(var->cost, -1.0))
         fprintf(fp, " - %s", tmp);
      else
         fprintf(fp, " %+.15g %s", var->cost, tmp);
      
      if (++cnt % 6 == 0)
         fprintf(fp, "\n ");
   }
   /* ---------------------------------------------------------------------- */

   /* First loop run for normal constraints, secound one for
    * lazy constraints, if any.
    */
   for(i = 0; i < 2; i++)
   {      
      fprintf(fp, "\n%s\n", (i == 0) ? "Subject to" : "Lazy Constraints");

      for(con = lp->con_root; con != NULL; con = con->next)
      {
         if (con->size == 0)
            continue;

         if ((con->flags & LP_FLAG_CON_SEPAR) == LP_FLAG_CON_SEPAR)
         {
            separate = TRUE;
            if (i == 0)
               continue;
         }
         else if (i == 1)
            continue;
         
         lps_makename(tmp, MPS_NAME_LEN + 1, con->name, con->number);
         fprintf(fp, " %s:\n ", tmp);

         for(nzo = con->first, cnt = 0; nzo != NULL; nzo = nzo->con_next)
         {
            lps_makename(tmp, MPS_NAME_LEN + 1,
               nzo->var->name, nzo->var->number);

            if (EQ(nzo->value, 1.0))
               fprintf(fp, " + %s", tmp);
            else if (EQ(nzo->value, -1.0))
               fprintf(fp, " - %s", tmp);
            else
               fprintf(fp, " %+.15g %s", nzo->value, tmp);
            
            if (++cnt % 6 == 0)
               fprintf(fp, "\n ");         
         }
         switch(con->type)
         {
         case CON_LE :
            fprintf(fp, " <= ");
            break;
         case CON_GE :
            fprintf(fp, " >= ");
            break;
         case CON_EQ :
            fprintf(fp, " = ");
            break;
         default :
            abort();
         }
         fprintf(fp, "%.15g\n", con->rhs);
      }
      if (!separate)
         break;
   }

   /* ---------------------------------------------------------------------- */

   fprintf(fp, "Bounds\n");

   for(var = lp->var_root; var != NULL; var = var->next)
   {
      if (var->size > 0)
      {
         if (LE(var->lower, -INFINITY))
            fprintf(fp, " -Inf");
         else
            fprintf(fp, " %.15g", var->lower);

         lps_makename(tmp, MPS_NAME_LEN + 1, var->name, var->number);
         fprintf(fp, " <= %s <= ", tmp);
         
         if (GE(var->upper, INFINITY))
            fprintf(fp, "+Inf\n");
         else
            fprintf(fp, "%.15g\n", var->upper);
      }
   }
   for(first = TRUE, var = lp->var_root; var != NULL; var = var->next)
   {
      if ((var->size > 0) && (var->type == VAR_INT))
      {
         if (first)
         {
            fprintf(fp, "General\n");
            first = FALSE;
         }
         lps_makename(tmp, MPS_NAME_LEN + 1, var->name, var->number);
         fprintf(fp, " %s\n", tmp);
      }
   }
   for(first = TRUE, var = lp->var_root; var != NULL; var = var->next)
   {
      if ((var->size > 0) && (var->type == VAR_BIN))
      {
         if (first)
         {
            fprintf(fp, "Binary\n");
            first = FALSE;
         }
         lps_makename(tmp, MPS_NAME_LEN + 1, var->name, var->number);
         fprintf(fp, " %s\n", tmp);
      }
   }
   fprintf(fp, "End\n");
}   




